import numpy as np
import random

def minkowski_dist(v1, v2, p):
    #Assumes v1 and v2 are equal length arrays of numbers
    dist = 0
    for i in range(len(v1)):
        dist += abs(v1[i] - v2[i])**p
    return dist**(1/p)

def scale_attrs(vals):
    vals = np.array(vals)
    mean = sum(vals)/len(vals)
    sd = np.std(vals)
    vals = vals - mean
    return vals/sd

def accuracy(true_pos, false_pos, true_neg, false_neg):
    numerator = true_pos + true_neg
    denominator = true_pos + true_neg + false_pos + false_neg
    return numerator/denominator

def sensitivity(true_pos, false_neg):
    try:
        return true_pos/(true_pos + false_neg)
    except ZeroDivisionError:
        return float('nan')
    
def specificity(true_neg, false_pos):
    try:
        return true_neg/(true_neg + false_pos)
    except ZeroDivisionError:
        return float('nan')
    
def pos_pred_val(true_pos, false_pos):
    try:
        return true_pos/(true_pos + false_pos)
    except ZeroDivisionError:
        return float('nan')
    
def neg_pred_val(true_neg, false_neg):
    try:
        return true_neg/(true_neg + false_neg)
    except ZeroDivisionError:
        return float('nan')
       
def get_stats(true_pos, false_pos, true_neg, false_neg):
    accur = accuracy(true_pos, false_pos, true_neg, false_neg)
    sens = sensitivity(true_pos, false_neg)
    spec = specificity(true_neg, false_pos)
    ppv = pos_pred_val(true_pos, false_pos)
    return (accur, sens, spec, ppv)


class Passenger(object):
    feature_names = ('C1', 'C2', 'C3', 'age', 'male gender')
    def __init__(self, p_class, age, gender, survived, name):
        self.name = name
        self.feature_vec = [0, 0, 0, age, gender]
        self.feature_vec[p_class - 1] = 1
        self.label = survived
        self.cabin_class = p_class
    def distance(self, other):
        return minkowski_dist(self.feature_vec, other.feature_vec, 2)
    def get_class(self):
        return self.cabin_class
    def get_age(self):
        return self.feature_vec[3]
    def get_gender(self):
        return self.feature_vec[4]
    def get_name(self):
        return self.name
    def get_features(self):
        return self.feature_vec[:]
    def get_label(self):
        return self.label

def get_Titanic_data(fname, to_scale_age = False):
    data = {}
    data['class'], data['survived'], data['age'] = [], [], []
    data['gender'], data['name'] = [], []
    f = open(fname)
    line = f.readline()
    while line != '':
        split = line.split(',')
        data['class'].append(int(split[0]))
        data['age'].append(float(split[1]))
        if split[2] == 'M': 
            data['gender'].append(1)
        else:
            data['gender'].append(0)
        if split[3] == '1':
            data['survived'].append('Survived')
        else:
            data['survived'].append('Died')
        data['name'].append(split[4:])
        line = f.readline()
    if to_scale_age:
        data['age'] = scale_attrs(data['age'])
    return data
                
def build_titanic_examples(find_name, to_scale = False):
    data = get_Titanic_data(find_name, to_scale)
    examples = []
    for i in range(len(data['class'])):
        p = Passenger(data['class'][i], data['age'][i],
                      data['gender'][i], data['survived'][i],
                      data['name'][i])
        examples.append(p)
    print('Finished processing', len(examples),
          'passengers\nwith to_scale =', to_scale, '\n')    
    return examples

def find_nearest(name, example_set, metric):
    for e in example_set:
        if e.get_name() == name:
            example = e
            break
    cur_dist = None
    for e in example_set:
        if e.get_name() != name:
            if cur_dist == None or\
               metric(example, e) < cur_dist:
                nearest = e
                cur_dist = metric(example, nearest)
    return nearest
   
def find_k_nearest(example, example_set, k):
    k_nearest, distances = [], []
    #Build lists containing first k examples and their distances
    for i in range(k):
        k_nearest.append(example_set[i])
        distances.append(example.distance(example_set[i]))
    max_dist = max(distances) #Get maximum distance
    #Look at examples not yet considered
    for e in example_set[k:]:
        dist = example.distance(e)
        if dist < max_dist:
            #replace farther neighbor by this one
            max_index = distances.index(max_dist)
            k_nearest[max_index] = e
            distances[max_index] = dist
            max_dist = max(distances)      
    return k_nearest, distances
    
def k_nearest_classify(training, test_set, label, k):
    """Assumes training & test_set lists of examples, k an int
       Predicts whether each example in test_set has label
       Returns number of true positives, false positives,
          true negatives, and false negatives"""
    true_pos, false_pos, true_neg, false_neg = 0, 0, 0, 0
    for test_case in test_set:
        nearest, distances = find_k_nearest(test_case, training, k)
        #conduct vote
        numMatch = 0
        for i in range(len(nearest)):
            if nearest[i].get_label() == label:
                numMatch += 1
        if numMatch > k//2: #guess label
            if test_case.get_label() == label:
                true_pos += 1
            else:
                false_pos += 1
        else: #guess not label
            if test_case.get_label() != label:
                true_neg += 1
            else:
                false_neg += 1
    return true_pos, false_pos, true_neg, false_neg

def split80_20(examples):
    sample_indices = random.sample(range(len(examples)),
                                  len(examples)//5)
    training_set, test_set = [], []
    for i in range(len(examples)):
        if i in sample_indices:
            test_set.append(examples[i])
        else:
            training_set.append(examples[i])
    return training_set, test_set
    
def random_splits(examples, method, num_splits, to_print = True):
    true_pos, false_pos, true_neg, false_neg = 0, 0, 0, 0
    for t in range(num_splits):
        training_set, test_set = split80_20(examples)
        results = method(training_set, test_set)
        true_pos += results[0]
        false_pos += results[1]
        true_neg += results[2]
        false_neg += results[3]
    return true_pos/num_splits, false_pos/num_splits,\
              true_neg/num_splits, false_neg/num_splits

def classify(examples, num_splits, n):
    knn = lambda training, test_set:\
                k_nearest_classify(training, test_set,
                                  'Survived', n)
    true_pos, false_pos, true_neg, false_neg =\
          random_splits(examples, knn, num_splits)
    
    return get_stats(true_pos, false_pos, true_neg, false_neg)
